from dataclasses import dataclass
from typing import Any, Callable, List, Optional, Union, overload

from dftracer.python.ai_common import DFTracerAI, dftracer
from dftracer.python.common import DFTRACER_ENABLE, P, R, TagDType, TagType, TagValue

try:
    import torch

    TORCH_AVAILABLE = True
except ImportError:
    TORCH_AVAILABLE = False
    torch = None  # type: ignore

if DFTRACER_ENABLE and TORCH_AVAILABLE:
    try:
        from functorch.compile import make_boxed_func

        # Alpha feature from: https://docs.pytorch.org/docs/stable/torch.compiler_custom_backends.html#custom-backends-after-aotautograd
        from torch._dynamo.backends.common import aot_autograd
    except ImportError:
        raise RuntimeError(
            "DFTracer dynamo requires PyTorch and functorch to be installed"
        )

CAT_DYNAMO = "dynamo"


@dataclass
class CallStackRecord:
    op_name: str
    op_type: str
    target: str
    timestamp_us: int
    grad_enabled: bool


def create_detailed_op_name(node: Any, gm: Any) -> str:
    """Create detailed operation name with layer/subfunction info"""
    base_name = str(node.target).replace(".", "_")

    if node.op == "call_module":
        module_path = str(node.target)
        module = gm.get_submodule(node.target)
        module_type = type(module).__name__
        return f"{module_path}.{module_type}_{base_name}"
    elif node.op == "call_function":
        func_name = getattr(node.target, "__name__", str(node.target))
        return f"function.{func_name}"
    elif node.op == "call_method":
        return f"method.{base_name}"
    else:
        return f"{node.op}.{base_name}"


class Dynamo(DFTracerAI):
    def __init__(
        self,
        name: Optional[str] = None,
        epoch: Optional[int] = None,
        step: Optional[int] = None,
        image_idx: Optional[int] = None,
        image_size: Optional[Any] = None,
        enable: bool = True,
    ):
        super().__init__(
            cat=CAT_DYNAMO,
            name=name,
            epoch=epoch,
            step=step,
            image_idx=image_idx,
            image_size=image_size,
            enable=enable,
        )
        self.call_stack: List[CallStackRecord] = []

    def reset(self) -> None:
        """Reset the tracer state"""
        super().reset()
        self.call_stack.clear()

    def _record_time_start(self, op_name: str, op_type: str, target: str) -> None:
        timestamp_us = self.get_time()
        grad_enabled = False
        if TORCH_AVAILABLE:
            grad_enabled = torch.is_grad_enabled()
        record = CallStackRecord(
            op_name=op_name,
            op_type=op_type,
            target=target,
            timestamp_us=timestamp_us,
            grad_enabled=grad_enabled,
        )
        self.call_stack.append(record)

    def _record_time_end(self, op_name: str, node: Any) -> None:
        """Record end time for an operation - called from graph execution"""
        if not self.call_stack:
            return None

        # The start_info is at the top of the call stack
        start_info = self.call_stack.pop()
        end_timestamp_us = self.get_time()

        # Calculate duration
        duration_us = end_timestamp_us - start_info.timestamp_us

        # Log the event
        dft = dftracer.get_instance()
        dft.enter_event()
        dft.log_event(
            name=op_name,
            cat=self.cat,
            start_time=start_info.timestamp_us,
            duration=duration_us,
            int_args={
                "grad_enabled": TagValue(
                    int(start_info.grad_enabled), TagDType.INT, TagType.KEY
                ).value(),
                # TODO: Add input and output shapes
            },
        )
        dft.exit_event()
        return None

    def add_enhanced_timing_to_graph(self, gm: Any) -> Any:
        """Add enhanced timing nodes to the FX graph"""
        graph = gm.graph

        # Insert timing calls before and after each operation
        nodes_to_instrument = []
        for node in graph.nodes:
            # We do not want to profile placeholder nodes
            if node.op in ["call_module", "call_function", "call_method"]:
                nodes_to_instrument.append(node)

        for node in reversed(nodes_to_instrument):
            op_name = create_detailed_op_name(node, gm)

            with graph.inserting_before(node):
                graph.call_function(
                    self._record_time_start,
                    args=(op_name, node.op, str(node.target)),
                )

            with graph.inserting_after(node):
                graph.call_function(self._record_time_end, args=(op_name, node))

        gm.recompile()
        # Print graph
        # gm.print_readable()
        return gm

    @overload
    def compile(
        self, f_py: Callable[P, R], autograd: bool = True
    ) -> Callable[P, R]: ...

    @overload
    def compile(
        self, f_py: None = None, *, autograd: bool = True
    ) -> Callable[[Callable[P, R]], Callable[P, R]]: ...

    def compile(
        self, f_py: Optional[Callable[P, R]] = None, autograd: bool = True
    ) -> Union[Callable[P, R], Callable[[Callable[P, R]], Callable[P, R]]]:
        """Used to compile torch.nn.Module and the forward functions
        - Uses PyTorch Dynamo to add nodes before and after each operation
        - The nodes are run before and after each operation -> providing detailed timing information
        """
        if not (DFTRACER_ENABLE and self.enable):  # type: ignore[truthy-function]
            return f_py  # type: ignore

        def _decorator(func: Callable[P, R]) -> Callable[P, R]:
            def enhanced_trace_wrapper(gm: Any, example: Any, **kwargs: Any) -> Any:
                """Enhanced custom trace function for Dynamo"""
                instrumented_gm = self.add_enhanced_timing_to_graph(gm)
                if autograd:
                    return make_boxed_func(instrumented_gm.forward)
                else:
                    return instrumented_gm.forward

            if autograd:
                return torch.compile(  # type: ignore
                    func, backend=aot_autograd(fw_compiler=enhanced_trace_wrapper)
                )
            else:
                return torch.compile(func, backend=enhanced_trace_wrapper)  # type: ignore

        return _decorator(f_py) if callable(f_py) else _decorator

    def count_parameters(self, module: Any) -> Optional[int]:
        """Count parameters in a module"""
        if hasattr(module, "parameters"):
            return int(sum(p.numel() for p in module.parameters()))
        return None


dynamo = Dynamo()

__all__ = ["Dynamo", "dynamo"]



from dftracer.python.env import *

logger = setup_stream_logger()

import dftracer.python.common as common

if DFTRACER_ENABLE:
    try:
        import dftracer.dftracer_dbg as profiler

        common.profiler = profiler
        logger.info("DFTracer was enabled.")
    except ImportError:
        logger.info("DFTracer was enabled but not available.")
        pass

from dftracer.python.common import *



import functools
import sys
from typing import Any, Callable, Dict, Iterator, List, Optional, overload

from dftracer.python.common import *
from dftracer.python.env import *

dft_fn = None  # type: ignore
dftracer = None  # type: ignore

if sys.version_info >= (3, 11):
    from enum import StrEnum as StringEnum
    from enum import auto
    # StrEnum in 3.11 already lower-case
else:
    from enum import Enum, auto

    # MIT License: https://github.com/irgeek/StrEnum/blob/master/strenum/__init__.py
    class StringEnum(str, Enum):
        # this is a specific version that will lowercase the values

        def __new__(cls, value: Any, *args: Any, **kwargs: Any) -> "StringEnum":
            if not isinstance(value, (str, auto)):
                raise TypeError(  # pragma: no cover
                    f"Values of StrEnums must be strings: {value!r} is a {type(value)}"
                )
            return super().__new__(cls, value, *args, **kwargs)

        def __str__(self) -> str:
            return str(self.value)

        @staticmethod
        def _generate_next_value_(
            name: str, start: int, count: int, last_values: List[Any]
        ) -> str:
            return name.lower()


ITER_COUNT_NAME = "count"
INIT_NAME = "init"
BLOCK_NAME = "block"
ITER_NAME = "iter"
CTX_SEPARATOR = "."
ROOT_NAME = "ai_root"
ROOT_CAT = "ai_root"
START_METADATA_NAME = "start"
STOP_METADATA_NAME = "end"


def get_iter_block_name(name: str) -> str:
    return (
        f"{name}{CTX_SEPARATOR}{BLOCK_NAME}"
        if not name.endswith(f"{CTX_SEPARATOR}{BLOCK_NAME}")
        else name
    )


def get_iter_handle_name(name: str) -> str:
    return (
        f"{name}{CTX_SEPARATOR}{ITER_NAME}"
        if not name.endswith(f"{CTX_SEPARATOR}{ITER_NAME}")
        else name
    )


class _DFTracerAI:
    def __init__(
        self,
        cat: str,
        name: Optional[str] = None,
        epoch: Optional[int] = None,
        step: Optional[int] = None,
        image_idx: Optional[int] = None,
        image_size: Optional[Any] = None,
        enable: bool = True,
    ) -> None:
        if not name:
            name = cat
        global dft_fn
        self.profiler = dft_fn(
            cat=cat,
            name=name,
            epoch=epoch,
            step=step,
            image_idx=image_idx,
            image_size=image_size,
            enable=enable,
        )

    @overload
    def __call__(
        self,
        fn: Callable[P, R],
        *,
        enable: Optional[bool] = None,
        epoch: Optional[int] = None,
        step: Optional[int] = None,
        image_idx: Optional[int] = None,
        image_size: Optional[Any] = None,
        args: Optional[Dict[str, Any]] = None,
    ) -> Callable[P, R]: ...

    @overload
    def __call__(
        self,
        *,
        enable: Optional[bool] = None,
        epoch: Optional[int] = None,
        step: Optional[int] = None,
        image_idx: Optional[int] = None,
        image_size: Optional[Any] = None,
        args: Optional[Dict[str, Any]] = None,
    ) -> "DFTracerAI": ...

    def __call__(
        self,
        fn: Optional[Callable[P, R]] = None,
        *,
        enable: Optional[bool] = None,
        epoch: Optional[int] = None,
        step: Optional[int] = None,
        image_idx: Optional[int] = None,
        image_size: Optional[Any] = None,
        args: Optional[Dict[str, Any]] = None,
    ) -> Union[Callable[P, R], "DFTracerAI"]:
        if epoch is not None:
            self.profiler._arguments_int["epoch"] = TagValue(
                epoch, TagDType.INT, TagType.KEY
            ).value()
        if step is not None:
            self.profiler._arguments_int["step"] = TagValue(
                step, TagDType.INT, TagType.KEY
            ).value()
        if image_idx is not None:
            self.profiler._arguments_int["image_idx"] = TagValue(
                image_idx, TagDType.INT, TagType.KEY
            ).value()
        if image_size is not None:
            self.profiler._arguments_float["image_size"] = TagValue(
                image_size, TagDType.FLOAT, TagType.KEY
            ).value()
        args = args or {}
        for key, value in args.items():
            if isinstance(value, TagValue):
                new_value = value._value
            else:
                new_value = value
            if isinstance(new_value, int):
                self.profiler._arguments_int[key] = TagValue(
                    new_value, TagDType.INT, TagType.KEY
                ).value()
            elif isinstance(new_value, float):
                self.profiler._arguments_float[key] = TagValue(
                    new_value, TagDType.FLOAT, TagType.KEY
                ).value()
            else:
                self.profiler._arguments_string[key] = TagValue(
                    str(new_value), TagDType.STRING, TagType.KEY
                ).value()

        is_enabled = self.profiler._enable if enable is None else enable

        if fn:

            def _decorator(f: Callable[P, R]) -> Callable[P, R]:
                @functools.wraps(f)
                def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
                    if is_enabled:
                        with self:
                            return f(*args, **kwargs)
                    return f(*args, **kwargs)

                return wrapper

            return _decorator(fn)
        else:
            return DFTracerAI(
                cat=self.profiler._cat,
                name=self.profiler._name,
                epoch=epoch,
                step=step,
                image_idx=image_idx,
                image_size=image_size,
                enable=is_enabled,
            )

    def __enter__(self) -> "_DFTracerAI":
        self.profiler.__enter__()
        # Reset flush state to ensure proper event logging on each context manager entry.
        # The underlying DFTracer logger was designed for one-time use objects, but this
        # class acts as a singleton. Without this reset, events won't flush after the
        # first __exit__ call due to DFTracer's internal _flush state management.
        self.profiler._flush = False
        return self

    def get_time(self) -> int:
        global dftracer
        return dftracer.get_instance().get_time()

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.profiler.__exit__(exc_type, exc_val, exc_tb)

    def start(self, metadata: bool = False) -> None:
        global dftracer
        if metadata:
            time = dftracer.get_instance().get_time()
            self.profiler.log_metadata(
                key=f"{self.profiler._name}{CTX_SEPARATOR}{START_METADATA_NAME}",
                value=str(time),
            )
        else:
            self.__enter__()

    def stop(self, metadata: bool = False) -> None:
        global dftracer
        if metadata:
            time = dftracer.get_instance().get_time()
            self.profiler.log_metadata(
                key=f"{self.profiler._name}{CTX_SEPARATOR}{STOP_METADATA_NAME}",
                value=str(time),
            )
        else:
            self.__exit__(None, None, None)

    def reset(self) -> None:
        self.profiler.reset()

    def enable(self) -> None:
        self.profiler._enable = True

    def disable(self) -> None:
        self.profiler._enable = False

    @property
    def cat(self) -> str:
        return self.profiler._cat

    @property
    def name(self) -> str:
        return self.profiler._name

    def update(
        self,
        epoch: Optional[int] = None,
        step: Optional[int] = None,
        image_idx: Optional[int] = None,
        image_size: Optional[Any] = None,
        args: Optional[Dict[str, Any]] = None,
    ) -> "_DFTracerAI":
        if args is None:
            args = {}
        if DFTRACER_ENABLE and self.profiler._enable:
            if epoch is not None:
                self.profiler._arguments_int["epoch"] = TagValue(
                    epoch, TagDType.INT, TagType.KEY
                ).value()
            if step is not None:
                self.profiler._arguments_int["step"] = TagValue(
                    step, TagDType.INT, TagType.KEY
                ).value()
            if image_idx is not None:
                self.profiler._arguments_int["image_idx"] = TagValue(
                    image_idx, TagDType.INT, TagType.KEY
                ).value()
            if image_size is not None:
                self.profiler._arguments_float["image_size"] = TagValue(
                    image_size, TagDType.FLOAT, TagType.KEY
                ).value()
            for key, value in args.items():
                if isinstance(value, TagValue):
                    new_value = value._value
                else:
                    new_value = value
                if isinstance(new_value, int):
                    self.profiler._arguments_int[key] = TagValue(
                        new_value, TagDType.INT, TagType.KEY
                    ).value()
                elif isinstance(new_value, float):
                    self.profiler._arguments_float[key] = TagValue(
                        new_value, TagDType.FLOAT, TagType.KEY
                    ).value()
                else:
                    self.profiler._arguments_string[key] = TagValue(
                        str(new_value), TagDType.STRING, TagType.KEY
                    ).value()
        return self

    def iter(
        self,
        iterator: Iterator,
        *,
        include_block: bool = True,
        include_iter: bool = True,
        iter_name: Optional[str] = None,
        block_name: Optional[str] = None,
    ) -> Iterator:
        global dftracer
        iter_name = iter_name or get_iter_handle_name(self.profiler._name)
        block_name = block_name or get_iter_block_name(self.profiler._name)
        iter_val = 1

        start: int = 0
        if DFTRACER_ENABLE and self.profiler._enable:
            start = dftracer.get_instance().get_time()

        for v in iterator:
            if DFTRACER_ENABLE and self.profiler._enable:
                end = dftracer.get_instance().get_time()
                t0 = dftracer.get_instance().get_time()

            yield v

            if DFTRACER_ENABLE and self.profiler._enable:
                t1 = dftracer.get_instance().get_time()
                self.profiler._arguments_int[ITER_COUNT_NAME] = TagValue(
                    iter_val, TagDType.INT, TagType.KEY
                ).value()
                if include_iter:
                    dftracer.get_instance().enter_event()
                    dftracer.get_instance().log_event(
                        name=iter_name,
                        cat=self.profiler._cat,
                        start_time=start,
                        duration=end - start,
                        int_args=self.profiler._arguments_int,
                        float_args=self.profiler._arguments_float,
                        string_args=self.profiler._arguments_string,
                    )
                    dftracer.get_instance().exit_event()

                if include_block:
                    dftracer.get_instance().enter_event()
                    dftracer.get_instance().log_event(
                        name=block_name,
                        cat=self.profiler._cat,
                        start_time=t0,
                        duration=t1 - t0,
                        int_args=self.profiler._arguments_int,
                        float_args=self.profiler._arguments_float,
                        string_args=self.profiler._arguments_string,
                    )
                    dftracer.get_instance().exit_event()

                iter_val += 1
                start = dftracer.get_instance().get_time()


class DFTracerAI(_DFTracerAI):
    def __init__(
        self,
        *,
        cat: str,
        name: Optional[str] = None,
        epoch: Optional[int] = None,
        step: Optional[int] = None,
        image_idx: Optional[int] = None,
        image_size: Optional[Any] = None,
        enable: bool = True,
    ):
        super().__init__(
            cat=cat,
            name=name,
            epoch=epoch,
            step=step,
            image_idx=image_idx,
            image_size=image_size,
            enable=enable,
        )
        self._children: Dict[str, DFTracerAI] = {}

    def create_children(self, names: Dict[str, str]) -> None:
        for attr, name in names.items():
            epoch_tuple = self.profiler._arguments_int.get("epoch")
            step_tuple = self.profiler._arguments_int.get("step")
            image_idx_tuple = self.profiler._arguments_int.get("image_idx")
            image_size_tuple = self.profiler._arguments_float.get("image_size")

            tracer = DFTracerAI(
                cat=self.profiler._cat,
                name=name,
                epoch=int(epoch_tuple[1]) if epoch_tuple else None,
                step=int(step_tuple[1]) if step_tuple else None,
                image_idx=int(image_idx_tuple[1]) if image_idx_tuple else None,
                image_size=float(image_size_tuple[1]) if image_size_tuple else None,
                enable=self.profiler._enable,
            )
            setattr(self, attr, tracer)
            self._children[attr] = tracer

    def update(
        self,
        epoch: Optional[int] = None,
        step: Optional[int] = None,
        image_idx: Optional[int] = None,
        image_size: Optional[Any] = None,
        args: Optional[Dict[str, Any]] = None,
    ) -> "DFTracerAI":
        super().update(
            epoch=epoch,
            step=step,
            image_idx=image_idx,
            image_size=image_size,
            args=args,
        )
        for tracer in self._children.values():
            tracer.update(
                epoch=epoch,
                step=step,
                image_idx=image_idx,
                image_size=image_size,
                args=args,
            )
        return self

    def enable(self) -> None:
        super().enable()
        for tracer in self._children.values():
            tracer.enable()

    def disable(self) -> None:
        super().disable()
        for tracer in self._children.values():
            tracer.disable()

    def derive(self, name: str) -> "DFTracerAI":
        _name = f"{self.profiler._name}{CTX_SEPARATOR}{name}"
        if _name in self._children:
            return self._children[_name]

        epoch_tuple = self.profiler._arguments_int.get("epoch")
        step_tuple = self.profiler._arguments_int.get("step")
        image_idx_tuple = self.profiler._arguments_int.get("image_idx")
        image_size_tuple = self.profiler._arguments_float.get("image_size")

        child = DFTracerAI(
            cat=self.profiler._cat,
            name=_name,
            epoch=int(epoch_tuple[1]) if epoch_tuple else None,
            step=int(step_tuple[1]) if step_tuple else None,
            image_idx=int(image_idx_tuple[1]) if image_idx_tuple else None,
            image_size=float(image_size_tuple[1]) if image_size_tuple else None,
            enable=self.profiler._enable,
        )
        self._children[_name] = child
        return child

    @overload
    def init(
        self,
        fn: Callable[P, R],
        *,
        enable: Optional[bool] = None,
        epoch: Optional[int] = None,
        step: Optional[int] = None,
        image_idx: Optional[int] = None,
        image_size: Optional[Any] = None,
        args: Optional[Dict[str, Any]] = None,
    ) -> Callable[P, R]: ...

    @overload
    def init(
        self,
        fn: None = None,
        *,
        enable: Optional[bool] = None,
        epoch: Optional[int] = None,
        step: Optional[int] = None,
        image_idx: Optional[int] = None,
        image_size: Optional[Any] = None,
        args: Optional[Dict[str, Any]] = None,
    ) -> "DFTracerAI": ...

    def init(
        self,
        fn: Optional[Callable[P, R]] = None,
        *,
        enable: Optional[bool] = None,
        epoch: Optional[int] = None,
        step: Optional[int] = None,
        image_idx: Optional[int] = None,
        image_size: Optional[Any] = None,
        args: Optional[Dict[str, Any]] = None,
    ) -> Union[Callable[P, R], "DFTracerAI"]:
        _name = f"{self.profiler._name}{CTX_SEPARATOR}{INIT_NAME}"
        if _name in self._children:
            return self._children[_name](  # type: ignore[misc]
                fn=fn,  # type: ignore[arg-type]
                enable=enable,
                epoch=epoch,
                step=step,
                image_idx=image_idx,
                image_size=image_size,
                args=args,
            )
        epoch_tuple = self.profiler._arguments_int.get("epoch")
        step_tuple = self.profiler._arguments_int.get("step")
        image_idx_tuple = self.profiler._arguments_int.get("image_idx")
        image_size_tuple = self.profiler._arguments_float.get("image_size")

        child = DFTracerAI(
            cat=self.profiler._cat,
            name=_name,
            epoch=int(epoch_tuple[1]) if epoch_tuple else None,
            step=int(step_tuple[1]) if step_tuple else None,
            image_idx=int(image_idx_tuple[1]) if image_idx_tuple else None,
            image_size=float(image_size_tuple[1]) if image_size_tuple else None,
            enable=self.profiler._enable,
        )
        self._children[_name] = child
        return child(  # type: ignore[misc]
            fn=fn,  # type: ignore[arg-type]
            enable=enable,
            epoch=epoch,
            step=step,
            image_idx=image_idx,
            image_size=image_size,
            args=args,
        )


# Enumerations


class ProfileCategory(StringEnum):
    COMPUTE = auto()
    DATA = auto()
    DATALOADER = auto()
    COMM = auto()
    DEVICE = auto()
    CHECKPOINT = auto()
    PIPELINE = auto()


class ComputeEvent(StringEnum):
    FORWARD = auto()
    BACKWARD = auto()
    STEP = auto()


class DataEvent(StringEnum):
    PREPROCESS = auto()
    ITEM = auto()


class DataLoaderEvent(StringEnum):
    FETCH = auto()


# Terminologies are taken from https://docs.pytorch.org/docs/stable/distributed.html
class CommunicationEvent(StringEnum):
    SEND = auto()
    RECEIVE = auto()
    BARRIER = auto()
    BCAST = auto()
    REDUCE = auto()
    ALL_REDUCE = auto()
    GATHER = auto()
    ALL_GATHER = auto()
    SCATTER = auto()
    REDUCE_SCATTER = auto()
    ALL_TO_ALL = auto()


class DeviceEvent(StringEnum):
    TRANSFER = auto()


class PipelineEvent(StringEnum):
    EPOCH = auto()
    TRAIN = auto()
    EVALUATE = auto()
    TEST = auto()


class CheckpointEvent(StringEnum):
    CAPTURE = auto()
    RESTART = auto()


class Compute(DFTracerAI):
    forward: DFTracerAI
    backward: DFTracerAI
    step: DFTracerAI

    def __init__(
        self,
        epoch: Optional[int] = None,
        step: Optional[int] = None,
        image_idx: Optional[int] = None,
        image_size: Optional[Any] = None,
        enable: bool = True,
    ):
        super().__init__(
            cat=ProfileCategory.COMPUTE,
            name=ProfileCategory.COMPUTE,
            epoch=epoch,
            step=step,
            image_idx=image_idx,
            image_size=image_size,
            enable=enable,
        )

        self.create_children(
            {
                "forward": ComputeEvent.FORWARD,
                "backward": ComputeEvent.BACKWARD,
                "step": ComputeEvent.STEP,
            }
        )


class Data(DFTracerAI):
    preprocess: DFTracerAI
    item: DFTracerAI

    def __init__(
        self,
        epoch: Optional[int] = None,
        step: Optional[int] = None,
        image_idx: Optional[int] = None,
        image_size: Optional[Any] = None,
        enable: bool = True,
    ):
        super().__init__(
            cat=ProfileCategory.DATA,
            name=ProfileCategory.DATA,
            epoch=epoch,
            step=step,
            image_idx=image_idx,
            image_size=image_size,
            enable=enable,
        )
        self.create_children(
            {
                "preprocess": DataEvent.PREPROCESS,
                "item": DataEvent.ITEM,
            }
        )


class DataLoader(DFTracerAI):
    fetch: DFTracerAI

    def __init__(
        self,
        epoch: Optional[int] = None,
        step: Optional[int] = None,
        image_idx: Optional[int] = None,
        image_size: Optional[Any] = None,
        enable: bool = True,
    ):
        super().__init__(
            cat=ProfileCategory.DATALOADER,
            name=ProfileCategory.DATALOADER,
            epoch=epoch,
            step=step,
            image_idx=image_idx,
            image_size=image_size,
            enable=enable,
        )
        self.create_children(
            {
                "fetch": DataLoaderEvent.FETCH,
            }
        )


class Communication(DFTracerAI):
    send: DFTracerAI
    receive: DFTracerAI
    barrier: DFTracerAI
    bcast: DFTracerAI
    reduce: DFTracerAI
    all_reduce: DFTracerAI
    gather: DFTracerAI
    all_gather: DFTracerAI
    scatter: DFTracerAI
    reduce_scatter: DFTracerAI
    all_to_all: DFTracerAI

    def __init__(
        self,
        epoch: Optional[int] = None,
        step: Optional[int] = None,
        image_idx: Optional[int] = None,
        image_size: Optional[Any] = None,
        enable: bool = True,
    ):
        super().__init__(
            cat=ProfileCategory.COMM,
            name=ProfileCategory.COMM,
            epoch=epoch,
            step=step,
            image_idx=image_idx,
            image_size=image_size,
            enable=enable,
        )
        self.create_children(
            {
                "send": CommunicationEvent.SEND,
                "receive": CommunicationEvent.RECEIVE,
                "barrier": CommunicationEvent.BARRIER,
                "bcast": CommunicationEvent.BCAST,
                "reduce": CommunicationEvent.REDUCE,
                "all_reduce": CommunicationEvent.ALL_REDUCE,
                "gather": CommunicationEvent.GATHER,
                "all_gather": CommunicationEvent.ALL_GATHER,
                "scatter": CommunicationEvent.SCATTER,
                "reduce_scatter": CommunicationEvent.REDUCE_SCATTER,
                "all_to_all": CommunicationEvent.ALL_TO_ALL,
            }
        )


class Device(DFTracerAI):
    transfer: DFTracerAI

    def __init__(
        self,
        epoch: Optional[int] = None,
        step: Optional[int] = None,
        image_idx: Optional[int] = None,
        image_size: Optional[Any] = None,
        enable: bool = True,
    ):
        super().__init__(
            cat=ProfileCategory.DEVICE,
            name=ProfileCategory.DEVICE,
            epoch=epoch,
            step=step,
            image_idx=image_idx,
            image_size=image_size,
            enable=enable,
        )
        self.create_children(
            {
                "transfer": DeviceEvent.TRANSFER,
            }
        )


class Checkpoint(DFTracerAI):
    capture: DFTracerAI
    restart: DFTracerAI

    def __init__(
        self,
        epoch: Optional[int] = None,
        step: Optional[int] = None,
        image_idx: Optional[int] = None,
        image_size: Optional[Any] = None,
        enable: bool = True,
    ):
        super().__init__(
            cat=ProfileCategory.CHECKPOINT,
            name=ProfileCategory.CHECKPOINT,
            epoch=epoch,
            step=step,
            image_idx=image_idx,
            image_size=image_size,
            enable=enable,
        )
        self.create_children(
            {
                "capture": CheckpointEvent.CAPTURE,
                "restart": CheckpointEvent.RESTART,
            }
        )


class Pipeline(DFTracerAI):
    epoch: DFTracerAI
    train: DFTracerAI
    evaluate: DFTracerAI
    test: DFTracerAI

    def __init__(
        self,
        epoch: Optional[int] = None,
        step: Optional[int] = None,
        image_idx: Optional[int] = None,
        image_size: Optional[Any] = None,
        enable: bool = True,
    ):
        super().__init__(
            cat=ProfileCategory.PIPELINE,
            name=ProfileCategory.PIPELINE,
            epoch=epoch,
            step=step,
            image_idx=image_idx,
            image_size=image_size,
            enable=enable,
        )
        self.create_children(
            {
                "epoch": PipelineEvent.EPOCH,
                "train": PipelineEvent.TRAIN,
                "evaluate": PipelineEvent.EVALUATE,
                "test": PipelineEvent.TEST,
            }
        )


# fmt: off
class AI(DFTracerAI):
    compute: Compute
    data: Data
    dataloader: DataLoader
    comm: Communication
    device: Device
    checkpoint: Checkpoint
    pipeline: Pipeline

    def __init__(
        self,
        epoch: Optional[int] = None,
        step: Optional[int] = None,
        image_idx: Optional[int] = None,
        image_size: Optional[Any] = None,
        enable: bool = True,
    ):
        super().__init__(cat=ROOT_CAT, name=ROOT_NAME, epoch=epoch, step=step, image_idx=image_idx, image_size=image_size, enable=enable)
        self.compute = Compute(epoch=epoch, step=step, image_idx=image_idx, image_size=image_size, enable=enable)
        self.data = Data(epoch=epoch, step=step, image_idx=image_idx, image_size=image_size, enable=enable)
        self.dataloader = DataLoader(epoch=epoch, step=step, image_idx=image_idx, image_size=image_size, enable=enable)
        self.comm = Communication(epoch=epoch, step=step, image_idx=image_idx, image_size=image_size, enable=enable)
        self.device = Device(epoch=epoch, step=step, image_idx=image_idx, image_size=image_size, enable=enable)
        self.checkpoint = Checkpoint(epoch=epoch, step=step, image_idx=image_idx, image_size=image_size, enable=enable)
        self.pipeline = Pipeline(epoch=epoch, step=step, image_idx=image_idx, image_size=image_size, enable=enable)

        self._children = {
            "compute": self.compute,
            "data": self.data,
            "dataloader": self.dataloader,
            "comm": self.comm,
            "device": self.device,
            "checkpoint": self.checkpoint,
            "pipeline": self.pipeline,
        }
# fmt: on



from dftracer.python.ai_common import *

ai = AI()
comm = ai.comm
compute = ai.compute
data = ai.data
dataloader = ai.dataloader
device = ai.device
checkpoint = ai.checkpoint
pipeline = ai.pipeline
checkpoint = ai.checkpoint

__all__ = [
    "INIT_NAME",
    "ITER_COUNT_NAME",
    "CommunicationEvent",
    "ComputeEvent",
    "DataEvent",
    "DataLoaderEvent",
    "DeviceEvent",
    "PipelineEvent",
    "ProfileCategory",
    "ai",
    "checkpoint",
    "comm",
    "compute",
    "data",
    "dataloader",
    "device",
    "get_iter_block_name",
    "get_iter_handle_name",
    "pipeline",
]



import inspect
import logging
import os
import signal
import sys
from collections.abc import Iterable, Sequence
from enum import Enum
from functools import wraps
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    Protocol,
    Tuple,
    TypeVar,
    Union,
    overload,
)

from dftracer.python.env import *
from typing_extensions import ParamSpec

P = ParamSpec("P")  # For capturing function parameters
R = TypeVar("R")  # For capturing function return types
T = TypeVar("T")  # For generic iterator types

# Type alias for TagValue return type
TagValueTuple = Tuple[int, Union[int, float, str]]


class ProfilerProtocol(Protocol):
    """Protocol defining the interface for profiler implementations."""

    def initialize(
        self,
        log_file: Optional[str] = None,
        data_dirs: Optional[str] = None,
        process_id: int = -1,
    ) -> None:
        """Initialize the profiler with configuration."""
        ...  # pragma: no cover

    def get_time(self) -> int:
        """Get current time for profiling."""
        ...  # pragma: no cover

    def enter_event(self) -> None:
        """Mark entry into an event."""
        ...  # pragma: no cover

    def exit_event(self) -> None:
        """Mark exit from an event."""
        ...  # pragma: no cover

    def log_event(
        self,
        name: str,
        cat: str,
        start_time: int,
        duration: int,
        int_args: Optional[Dict[str, TagValueTuple]] = None,
        string_args: Optional[Dict[str, TagValueTuple]] = None,
        float_args: Optional[Dict[str, TagValueTuple]] = None,
    ) -> None:
        """Log a profiling event."""
        ...  # pragma: no cover

    def log_metadata_event(self, key: str, value: str) -> None:
        """Log a metadata event."""
        ...  # pragma: no cover

    def finalize(self) -> None:
        """Finalize the profiler and release resources."""
        ...  # pragma: no cover


class NoOpProfiler:
    """Fallback no-operation profiler when pydftracer is not available."""

    def initialize(
        self,
        log_file: Optional[str] = None,
        data_dirs: Optional[str] = None,
        process_id: int = -1,
    ) -> None:
        pass

    def get_time(self) -> int:
        return 0

    def enter_event(self) -> None:
        pass

    def exit_event(self) -> None:
        pass

    def log_event(
        self,
        name: str,
        cat: str,
        start_time: int,
        duration: int,
        int_args: Optional[Dict[str, TagValueTuple]] = None,
        string_args: Optional[Dict[str, TagValueTuple]] = None,
        float_args: Optional[Dict[str, TagValueTuple]] = None,
    ) -> None:
        pass

    def log_metadata_event(self, key: str, value: str) -> None:
        pass

    def finalize(self) -> None:
        pass


# Profiler instance - will be set by logger.py during initialization
# Initialize with NoOpProfiler as default fallback
profiler: ProfilerProtocol = NoOpProfiler()


class TagType(Enum):
    KEY = 0
    VALUE = 1
    IGNORE = 2

    def __int__(self) -> int:
        return self.value


class TagDType(Enum):
    INT = 0
    FLOAT = 1
    STRING = 2

    def __int__(self) -> int:
        return self.value


class TagValue:
    def __init__(
        self,
        value: Any,
        dtype: TagDType = TagDType.STRING,
        tag_type: TagType = TagType.KEY,
    ) -> None:
        self._dtype = dtype
        self._value: Union[int, float, str]
        self._tag_type = tag_type
        if self._dtype == TagDType.INT:
            self._value = int(value)
        elif self._dtype == TagDType.FLOAT:
            self._value = float(value)
        else:
            self._value = str(value)

    def value(self) -> TagValueTuple:
        return (int(self._tag_type), self._value)


def capture_signal(signal_number: int, frame: Any) -> None:
    dftracer.get_instance().finalize()
    sys.exit(signal_number)


if DFTRACER_ENABLE:
    signal.signal(signal.SIGABRT, capture_signal)
    signal.signal(signal.SIGINT, capture_signal)
    signal.signal(signal.SIGTERM, capture_signal)


def setup_logger(
    name: str, log_file: str, formatter: str, level: int = logging.INFO
) -> logging.Logger:
    """To setup as many loggers as you want"""
    handler = logging.FileHandler(log_file)
    handler.setFormatter(logging.Formatter(formatter))
    logger = logging.getLogger(name)
    logger.setLevel(level)
    logger.addHandler(handler)
    return logger


class dftracer:
    __instance: Optional["dftracer"] = None

    def __init__(self) -> None:
        self.logger: Optional[ProfilerProtocol] = None
        self.dbg_logging: Optional[logging.Logger] = None
        dftracer.__instance = self

    @classmethod
    def get_instance(cls) -> "dftracer":
        """Static access method."""
        if dftracer.__instance is None:
            dftracer()
        # for linter (e.g. mypy) to understand that this won't be None
        assert dftracer.__instance is not None
        return dftracer.__instance

    @staticmethod
    def initialize_log(
        logfile: Optional[str] = None,
        data_dir: Optional[str] = None,
        process_id: int = -1,
    ) -> "dftracer":
        instance = dftracer.get_instance()
        global profiler
        log_file_path = None
        if logfile:
            log_file_path = Path(logfile)
        outfile = "dft.log"
        if DFTRACER_ENABLE:
            if log_file_path:
                os.makedirs(log_file_path.parent, exist_ok=True)
                outfile = os.path.join(log_file_path.parent, "dft.log")
        log_level = logging.ERROR
        if DFTRACER_LOG_LEVEL == "DEBUG":
            log_level = logging.DEBUG
        elif DFTRACER_LOG_LEVEL == "INFO":
            log_level = logging.INFO
        elif DFTRACER_LOG_LEVEL == "WARN":
            log_level = logging.WARN
        instance.dbg_logging = setup_logger(
            name="dftracer_dbg",
            log_file=outfile,
            formatter="[DFTRACER_PY %(levelname)s] %(message)s [%(pathname)s:%(lineno)d]",
            level=log_level,
        )
        instance.dbg_logging.debug(
            f"logger.initialize_log {logfile} {data_dir} {process_id}"
        )
        if DFTRACER_ENABLE:
            instance.logger = profiler
            instance.dbg_logging.debug(
                f"logger.initialize {logfile} {data_dir} {process_id}"
            )
            instance.logger.initialize(
                log_file=logfile, data_dirs=data_dir, process_id=process_id
            )
        return instance

    def get_time(self) -> int:
        if DFTRACER_ENABLE and self.logger:
            t = self.logger.get_time()
            if self.dbg_logging:
                self.dbg_logging.debug(f"logger.get_time {t}")
            return t
        return 0

    def enter_event(self) -> None:
        if DFTRACER_ENABLE and self.logger:
            self.logger.enter_event()
            if self.dbg_logging:
                self.dbg_logging.debug("logger.enter_event")

    def exit_event(self) -> None:
        if DFTRACER_ENABLE and self.logger:
            self.logger.exit_event()
            if self.dbg_logging:
                self.dbg_logging.debug("logger.exit_event")

    def log_event(
        self,
        name: str,
        cat: str,
        start_time: int,
        duration: int,
        int_args: Optional[Dict[str, TagValueTuple]] = None,
        float_args: Optional[Dict[str, TagValueTuple]] = None,
        string_args: Optional[Dict[str, TagValueTuple]] = None,
    ) -> None:
        if DFTRACER_ENABLE and self.logger:
            if self.dbg_logging:
                self.dbg_logging.debug(
                    f"logger.log_event {name} {cat} {start_time} {duration} int={int_args} str={string_args} float={float_args}"
                )
            self.logger.log_event(
                name=name,
                cat=cat,
                start_time=start_time,
                duration=duration,
                int_args=int_args or {},
                float_args=float_args or {},
                string_args=string_args or {},
            )

    def log_metadata_event(self, key: str, value: str) -> None:
        if DFTRACER_ENABLE and self.logger:
            if self.dbg_logging:
                self.dbg_logging.debug(f"logger.log_metadata_event {key} {value}")
            self.logger.log_metadata_event(key=key, value=value)

    def finalize(self) -> None:
        if DFTRACER_ENABLE and self.logger:
            if self.dbg_logging:
                self.dbg_logging.debug("logger.finalize")
            self.logger.finalize()


def get_default_args(func: Callable) -> Dict[str, Any]:
    signature = inspect.signature(func)
    return {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }


class dft_fn:
    def __init__(
        self,
        cat: str,
        name: Optional[str] = None,
        epoch: Optional[int] = None,
        step: Optional[int] = None,
        image_idx: Optional[int] = None,
        image_size: Optional[Any] = None,
        enable: bool = True,
    ) -> None:
        self._enable = enable
        self._cat = cat
        self._arguments_int: Dict[str, TagValueTuple] = {}
        self._arguments_float: Dict[str, TagValueTuple] = {}
        self._arguments_string: Dict[str, TagValueTuple] = {}
        self._t1: int = 0
        self._t2: int = 0
        self._flush: bool = False

        if not name:
            name = inspect.stack()[1].function
        self._name = name

        if DFTRACER_ENABLE and self._enable:
            if epoch is not None:
                self._arguments_int["epoch"] = TagValue(
                    epoch, TagDType.INT, TagType.KEY
                ).value()
            if step is not None:
                self._arguments_int["step"] = TagValue(
                    step, TagDType.INT, TagType.KEY
                ).value()
            if image_idx is not None:
                self._arguments_int["image_idx"] = TagValue(
                    image_idx, TagDType.INT, TagType.KEY
                ).value()
            if image_size is not None:
                self._arguments_float["image_size"] = TagValue(
                    image_size, TagDType.FLOAT, TagType.KEY
                ).value()
            self.reset()

    def __enter__(self) -> "dft_fn":
        if DFTRACER_ENABLE and self._enable:
            self._t1 = dftracer.get_instance().get_time()
            dftracer.get_instance().enter_event()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        if DFTRACER_ENABLE and self._enable:
            if not self._flush:
                self.flush()

    def update(
        self,
        epoch: Optional[int] = None,
        step: Optional[int] = None,
        image_idx: Optional[int] = None,
        image_size: Optional[Any] = None,
        args: Optional[Dict[str, Any]] = None,
    ) -> "dft_fn":
        if DFTRACER_ENABLE and self._enable:
            if epoch is not None:
                self._arguments_int["epoch"] = TagValue(
                    epoch, TagDType.INT, TagType.KEY
                ).value()
            if step is not None:
                self._arguments_int["step"] = TagValue(
                    step, TagDType.INT, TagType.KEY
                ).value()
            if image_idx is not None:
                self._arguments_int["image_idx"] = TagValue(
                    image_idx, TagDType.INT, TagType.KEY
                ).value()
            if image_size is not None:
                self._arguments_float["image_size"] = TagValue(
                    image_size, TagDType.FLOAT, TagType.KEY
                ).value()
            if args is not None:
                for key, value in args.items():
                    if isinstance(value, TagValue):
                        new_value = value._value
                    else:
                        new_value = value
                    if isinstance(new_value, int):
                        self._arguments_int[key] = TagValue(
                            new_value, TagDType.INT, TagType.KEY
                        ).value()
                    elif isinstance(new_value, float):
                        self._arguments_float[key] = TagValue(
                            new_value, TagDType.FLOAT, TagType.KEY
                        ).value()
                    else:
                        self._arguments_string[key] = TagValue(
                            str(new_value), TagDType.STRING, TagType.KEY
                        ).value()
        return self

    def flush(self) -> "dft_fn":
        if DFTRACER_ENABLE and self._enable:
            self._t2 = dftracer.get_instance().get_time()
            dftracer.get_instance().log_event(
                name=self._name,
                cat=self._cat,
                start_time=self._t1,
                duration=self._t2 - self._t1,
                int_args=self._arguments_int,
                float_args=self._arguments_float,
                string_args=self._arguments_string,
            )
            dftracer.get_instance().exit_event()
            self._flush = True
        return self

    def reset(self) -> "dft_fn":
        if DFTRACER_ENABLE and self._enable:
            self._t1 = dftracer.get_instance().get_time()
            dftracer.get_instance().enter_event()
            self._t2 = self._t1
            self._flush = False
        return self

    @overload
    def log(
        self, func: Callable[P, R], name: Optional[str] = None
    ) -> Callable[P, R]: ...

    @overload
    def log(
        self, func: None = None, *, name: Optional[str] = None
    ) -> Callable[[Callable[P, R]], Callable[P, R]]: ...

    def log(
        self, func: Optional[Callable[P, R]] = None, name: Optional[str] = None
    ) -> Union[Callable[P, R], Callable[[Callable[P, R]], Callable[P, R]]]:
        # CC BY-SA 4.0 https://stackoverflow.com/a/60832711
        def _decorator(f: Callable[P, R]) -> Callable[P, R]:
            _name = name if name else f.__qualname__
            arg_names: List[str] = []
            if DFTRACER_ENABLE and self._enable:
                arg_names = inspect.getfullargspec(f)[0]
                self._arguments_int = {}
                self._arguments_float = {}
                self._arguments_string = {}

            @wraps(f)
            def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
                if DFTRACER_ENABLE and self._enable:
                    if len(arg_names) > 0:
                        if "self" == arg_names[0]:
                            if hasattr(args[0], "epoch"):
                                self._arguments_int["epoch"] = TagValue(
                                    args[0].epoch, TagDType.INT, TagType.KEY
                                ).value()
                            if hasattr(args[0], "step"):
                                self._arguments_int["step"] = TagValue(
                                    args[0].step, TagDType.INT, TagType.KEY
                                ).value()
                            if hasattr(args[0], "image_size"):
                                self._arguments_float["image_size"] = TagValue(
                                    args[0].image_size, TagDType.FLOAT, TagType.KEY
                                ).value()
                            if hasattr(args[0], "image_idx"):
                                self._arguments_int["image_idx"] = TagValue(
                                    args[0].image_idx, TagDType.INT, TagType.KEY
                                ).value()
                        full_args = dict(zip(arg_names[1:], args[1:]))
                        full_args.update(kwargs)
                        full_args.update(get_default_args(f))

                        for param_name, value in full_args.items():
                            if param_name == "epoch":
                                self._arguments_int["epoch"] = TagValue(
                                    value, TagDType.INT, TagType.KEY
                                ).value()
                            elif param_name == "image_idx":
                                self._arguments_int["image_idx"] = TagValue(
                                    value, TagDType.INT, TagType.KEY
                                ).value()
                            elif param_name == "image_size":
                                self._arguments_float["image_size"] = TagValue(
                                    value, TagDType.FLOAT, TagType.KEY
                                ).value()
                            elif param_name == "step":
                                self._arguments_int["step"] = TagValue(
                                    value, TagDType.INT, TagType.KEY
                                ).value()

                    start = dftracer.get_instance().get_time()
                    dftracer.get_instance().enter_event()
                x = f(*args, **kwargs)
                if DFTRACER_ENABLE and self._enable:
                    end = dftracer.get_instance().get_time()
                    dftracer.get_instance().log_event(
                        name=_name,
                        cat=self._cat,
                        start_time=start,
                        duration=end - start,
                        int_args=self._arguments_int,
                        float_args=self._arguments_float,
                        string_args=self._arguments_string,
                    )
                    dftracer.get_instance().exit_event()
                return x

            return wrapper

        return _decorator(func) if callable(func) else _decorator

    def log_metadata(self, key: str, value: str) -> None:
        if DFTRACER_ENABLE and self._enable:
            dftracer.get_instance().log_metadata_event(key=key, value=value)

    def iter(
        self,
        func: Union[Sequence[T], Iterable[T], Generator[T, Any, Any]],
        name: str = "loop",
        iter_name: str = "step",
    ) -> Generator[T, Any, Any]:
        iter_val = 1
        _name = ""
        kernel_name = ""
        start = 0
        if DFTRACER_ENABLE and self._enable:
            _name = f"{name}.iter"
            kernel_name = f"{name}.yield"
            start = dftracer.get_instance().get_time()
            self._arguments_int = {}
            self._arguments_float = {}
            self._arguments_string = {}

        for v in func:
            if DFTRACER_ENABLE and self._enable:
                end = dftracer.get_instance().get_time()
                t0 = dftracer.get_instance().get_time()
            yield v
            if DFTRACER_ENABLE and self._enable:
                t1 = dftracer.get_instance().get_time()
                self._arguments_int[iter_name] = TagValue(
                    iter_val, TagDType.INT, TagType.KEY
                ).value()
                dftracer.get_instance().enter_event()
                dftracer.get_instance().log_event(
                    name=_name,
                    cat=self._cat,
                    start_time=start,
                    duration=end - start,
                    int_args=self._arguments_int,
                    float_args=self._arguments_float,
                    string_args=self._arguments_string,
                )
                dftracer.get_instance().exit_event()
                dftracer.get_instance().enter_event()
                dftracer.get_instance().log_event(
                    name=kernel_name,
                    cat=self._cat,
                    start_time=t0,
                    duration=t1 - t0,
                    int_args=self._arguments_int,
                    float_args=self._arguments_float,
                    string_args=self._arguments_string,
                )
                dftracer.get_instance().exit_event()

                iter_val += 1
                start = dftracer.get_instance().get_time()

    @overload
    def log_init(
        self, init: Callable[P, R], name: Optional[str] = None
    ) -> Callable[P, R]: ...

    @overload
    def log_init(
        self, init: None = None, *, name: Optional[str] = None
    ) -> Callable[[Callable[P, R]], Callable[P, R]]: ...

    def log_init(
        self, init: Optional[Callable[P, R]] = None, name: Optional[str] = None
    ) -> Union[Callable[P, R], Callable[[Callable[P, R]], Callable[P, R]]]:
        # CC BY-SA 4.0 https://stackoverflow.com/a/60832711
        def _decorator(fn: Callable[P, R]) -> Callable[P, R]:
            _name = name if name else fn.__qualname__
            arg_names: List[str] = []
            if DFTRACER_ENABLE and self._enable:
                arg_names = inspect.getfullargspec(fn)[0]
                self._arguments_int = {}
                self._arguments_float = {}
                self._arguments_string = {}

            @wraps(fn)
            def new_init(*args: Any, **kwargs: Any) -> Any:
                if DFTRACER_ENABLE and self._enable:
                    arg_values = dict(zip(arg_names, args))
                    arg_values.update(kwargs)
                    arg_values.update(get_default_args(fn))
                    if "epoch" in arg_values and arg_values["epoch"] is not None:
                        self._arguments_int["epoch"] = TagValue(
                            arg_values["epoch"], TagDType.INT, TagType.KEY
                        ).value()
                    if (
                        "image_idx" in arg_values
                        and arg_values["image_idx"] is not None
                    ):
                        self._arguments_int["image_idx"] = TagValue(
                            arg_values["image_idx"], TagDType.INT, TagType.KEY
                        ).value()
                    if (
                        "image_size" in arg_values
                        and arg_values["image_size"] is not None
                    ):
                        self._arguments_float["image_size"] = TagValue(
                            arg_values["image_size"], TagDType.FLOAT, TagType.KEY
                        ).value()
                    if "step" in arg_values and arg_values["step"] is not None:
                        self._arguments_int["step"] = TagValue(
                            arg_values["step"], TagDType.INT, TagType.KEY
                        ).value()
                    start = dftracer.get_instance().get_time()
                    dftracer.get_instance().enter_event()
                fn(*args, **kwargs)
                if DFTRACER_ENABLE and self._enable:
                    end = dftracer.get_instance().get_time()
                    dftracer.get_instance().log_event(
                        name=_name,
                        cat=self._cat,
                        start_time=start,
                        duration=end - start,
                        int_args=self._arguments_int,
                        float_args=self._arguments_float,
                        string_args=self._arguments_string,
                    )
                    dftracer.get_instance().exit_event()

            return new_init

        return _decorator(init) if callable(init) else _decorator

    @overload
    def log_static(
        self, func: Callable[P, R], name: Optional[str] = None
    ) -> Callable[P, R]: ...

    @overload
    def log_static(
        self, func: None = None, *, name: Optional[str] = None
    ) -> Callable[[Callable[P, R]], Callable[P, R]]: ...

    def log_static(
        self, func: Optional[Callable[P, R]] = None, name: Optional[str] = None
    ) -> Union[Callable[P, R], Callable[[Callable[P, R]], Callable[P, R]]]:
        # CC BY-SA 4.0 https://stackoverflow.com/a/60832711
        def _decorator(f: Callable[P, R]) -> Callable[P, R]:
            # Extract actual function if we received a staticmethod object
            if isinstance(f, staticmethod):
                actual_func = f.__func__
            else:
                actual_func = f

            _name = name if name else actual_func.__qualname__

            @wraps(actual_func)
            def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
                if DFTRACER_ENABLE and self._enable:
                    start = dftracer.get_instance().get_time()
                    dftracer.get_instance().enter_event()
                x = actual_func(*args, **kwargs)
                if DFTRACER_ENABLE and self._enable:
                    end = dftracer.get_instance().get_time()
                    dftracer.get_instance().log_event(
                        name=_name,
                        cat=self._cat,
                        start_time=start,
                        duration=end - start,
                        int_args=self._arguments_int,
                        float_args=self._arguments_float,
                        string_args=self._arguments_string,
                    )
                    dftracer.get_instance().exit_event()
                return x

            return wrapper

        return _decorator(func) if callable(func) else _decorator



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html


# Don't add project root to path - we want to use the installed package from site-packages
# If we add the project root, Python will try to import from source which may cause issues
# sys.path.insert(0, str(Path(__file__).parent.parent.parent))

# Mock imports for packages that may not be available during doc build
autodoc_mock_imports = []

# Try to import the package
try:
    import dftracer.python  # noqa: F401

    print("✓ dftracer.python package found and imported successfully.")
except (ImportError, ModuleNotFoundError) as e:
    print(f"Warning: dftracer.python package not found: {e}")
    print("API documentation will have limited information.")
    print("To generate full API docs, install the package: pip install -e .")
    # Don't mock - let it fail to show what's missing
    # autodoc_mock_imports = ['dftracer', 'dftracer.python']

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information


project = "pydftracer"
copyright = "%Y, Hariharan Devarajan, Ray Andrew Sinurat"
author = "Hariharan Devarajan, Ray Andrew Sinurat"

# The version info for the project
# Try to get version from the package
try:
    from importlib.metadata import version

    release = version("pydftracer")
    version = ".".join(release.split(".")[:2])
except Exception:
    version = "0.1"
    release = "0.1.0"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "myst_parser",  # For Markdown support
]

# Napoleon settings for Google/NumPy style docstrings
napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = True
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_preprocess_types = False
napoleon_type_aliases = None
napoleon_attr_annotations = True

# Add mappings for intersphinx - link to main DFTracer docs and Python docs
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "dftracer": ("https://dftracer.readthedocs.io/en/latest/", None),
}

templates_path = ["_templates"]
exclude_patterns = []

# The suffix(es) of source filenames.
source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}

# The master toctree document.
master_doc = "index"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]

# Theme options
html_theme_options = {
    "navigation_depth": 4,
    "collapse_navigation": False,
    "sticky_navigation": True,
    "includehidden": True,
    "titles_only": False,
}

# -- Options for autodoc -----------------------------------------------------
autodoc_default_options = {
    "members": True,
    "member-order": "bysource",
    "special-members": "__init__",
    "undoc-members": True,
    "exclude-members": "__weakref__",
}

# Use type stubs (.pyi files) for documentation
autodoc_typehints = "description"
autodoc_typehints_description_target = "documented"

# -- Options for todo extension ----------------------------------------------
todo_include_todos = True

# -- Options for autosummary -------------------------------------------------
autosummary_generate = True



import logging

from dftracer.python.env import *

logger: logging.Logger = setup_stream_logger()

import dftracer.python.common as common

if DFTRACER_ENABLE:
    try:
        import dftracer.dftracer as profiler  # type: ignore

        common.profiler = profiler  # type: ignore
        logger.info("DFTracer was enabled.")
    except ImportError:  # pragma: no cover
        logger.info("DFTracer was enabled but not available.")

from dftracer.python.common import *  # noqa: F403, F401, E402



import logging
import os

DFTRACER_ENABLE_ENV = "DFTRACER_ENABLE"
DFTRACER_INIT_ENV = "DFTRACER_INIT"
DFTRACER_LOG_LEVEL_ENV = "DFTRACER_LOG_LEVEL"

DFTRACER_ENABLE = True if os.getenv(DFTRACER_ENABLE_ENV, "0") == "1" else False
DFTRACER_INIT_PRELOAD = (
    True if os.getenv(DFTRACER_INIT_ENV, "PRELOAD") == "PRELOAD" else False
)
DFTRACER_LOG_LEVEL = os.getenv(DFTRACER_LOG_LEVEL_ENV, "ERROR")


def setup_stream_logger() -> logging.Logger:
    log_level = logging.ERROR
    if DFTRACER_LOG_LEVEL == "DEBUG":
        log_level = logging.DEBUG
    elif DFTRACER_LOG_LEVEL == "INFO":
        log_level = logging.INFO
    elif DFTRACER_LOG_LEVEL == "WARN":
        log_level = logging.WARN
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter("[%(levelname)s] %(message)s"))
    logger = logging.getLogger("DFTRACER")
    logger.setLevel(log_level)
    logger.addHandler(handler)
    return logger



import dftracer.python.ai_common as ai_common
from dftracer.python.logger import dft_fn as dft_fn_orig
from dftracer.python.logger import dftracer as dftracer_orig

ai_common.dft_fn = dft_fn_orig  # type: ignore[misc]
ai_common.dftracer = dftracer_orig  # type: ignore[misc]

from dftracer.python.ai_common import *
from dftracer.python.ai_init import *



from importlib.metadata import PackageNotFoundError, version

from dftracer.python.ai import *
from dftracer.python.dynamo import *
from dftracer.python.logger import *

try:
    __version__ = version("pydftracer")
except PackageNotFoundError:
    # package is not installed
    pass



# we presume installed build dependencies
from __future__ import annotations

from setuptools import setup
from setuptools_scm import ScmVersion


def myversion_func(version: ScmVersion) -> str:
    from setuptools_scm.version import only_version

    return version.format_next_version(only_version, fmt="{tag}.dev{distance}")


setup(use_scm_version={"version_scheme": myversion_func})
